#!/usr/local/bin/python

'''
Web Cacheability Engine
Copyright 1999 Mark Nottingham <mnot@pobox.com>

THIS SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND.

This software may be freely distributed, but may not be modified in any way.
See the LICENSE file for more information.

INSTALLATION NOTES
* This script MUST be run on a host that has an accurate clock; the best way
  to do this is with NTP <http://www.ntp.org/>

* Upstream proxies can have a profound effect on the results of this script;
  although it is possible to use one, it is not recommended. This also applies
  for transparent proxies; check to see if your service provider uses one.
  In particular, the Date, Expires and Last-Modified checks will be inaccurate.

* spider.py, httpheaders.py and httpheadertypes.py MUST be available either:
  - in the same directory as this script
  - in Python's site-packages directory (in lib/python1.5; you may have to 
    create it)

FUTURE DEVELOPMENTS
* Authentication support
* validation on ETags
* check for HEAD support
* Chunked Encoding Check
* threading of the spidering engine (persistent connections)
* HTML comments in description (good/bad, link to more info)

TODO
* script_URL needs to be Windows-friendly
* cookie printing / etc
'''


# This file was edited with four space tab stops. No tabs were harmed during 
# the production of this program.


import time, sys, os, types
from string import lower, join, split, strip, find
from urlparse import urlparse, urlunparse, urljoin
from urllib import quote
from spider import Spider, no_date, Error

__version__ = '1.30'
 
class HTMLSpider(Spider):
	''' HTML output version of the spider. '''

	def __init__(self, url):
		Spider.__init__(self, url)
		self.no_data = ' - '

	def __repr__(self):
		if type(self._errmsg) is types.InstanceType:
			self._errmsg = self._errmsg.args[-1]
		o=[]
		p = o.append
		try:
			description = self.write_description()
		except:
			description = ''

		if self.niceness < 0:
			niceness = 'red'
		elif self.niceness > 0:
			niceness = 'green'
		else:
			niceness = 'yellow'

		p("<LI class='%s'><H2><A href='%s' target='view'>%s</a> " % (
			niceness, self.request_url, self.request_url))
		if self._redirectlevel > 0:
			p(" &nbsp; <EM class='bad'>(redirects: %s)</EM>" % (self._redirectlevel))
		p("</H2>")
		if self._errmsg:
			p(" <EM>ERROR: %s</EM>" % (self._errmsg))
			return join(o, "\n")

		p("\n")
		p("<TABLE cellspacing=0 cellpadding=0>")
		
		if self.date_skew == None or \
		  abs(self.date_skew) > self.skew_tolerance * 60:
			p("<TR><TH>Date &nbsp;</TH>")
			try:
				raw_date = self._hdrs.date.raw
			except:
				raw_date = self.no_data
			p("<TD><SPAN class='red'> %s </SPAN></TD></TR>\n" % (raw_date))
		p("<TR><TH>Expires &nbsp;</TH>")
		if self._hdrs.has_key('Expires'):
			if self._hdrs.expires:
				e_date, e_rel = print_date(
				  self._hdrs.expires, self.date_fetched)
			else:
				e_rel = str(self._hdrs.expires.raw)
				e_date = 'immediately'
			p("<TD> %s &nbsp; (%s)</TD></TR>\n" % (e_rel, e_date))
		else:
			p("<TD>&nbsp; %s </TD></TR>\n" % (self.no_data))
		
		p("<TR><TH>Cache-Control &nbsp;</TH>")
		p("<TD> %s</TD></TR>\n" % (
			self._hdrs.get('Cache-Control', "&nbsp;" + self.no_data)))

		p("<TR><TH>Last-Modified &nbsp;</TH>")
		p("<TD><NOBR>")
		if self._hdrs.has_key('last-modified'):
			m_date, m_rel = print_date(
			  self._hdrs.last_modified, self.date_fetched)
			if not m_rel: 
				m_rel="<SPAN class='red'>invalid</SPAN>"
			p("%s &nbsp;(%s)" % (m_rel, self._hdrs.last_modified.raw))
		else:
			p("&nbsp;" + self.no_data)
		if self._validated == 0:
			p(" <EM class='bad'>validation returned same object</EM>")
		elif self._validated == 1:
			p(" <EM class='ok'>validated</EM>")
		p("</NOBR></TD></TR>\n")
			
		p("<TR><TH>ETag &nbsp;</TH>")
		p("<TD> %s</TD></TR>\n" % (self._hdrs.get('ETag', "&nbsp;" + self.no_data)))

		if self._hdrs.has_key('Set-Cookie'):
			p("<TR><TH>Set-Cookie &nbsp;</TH>")
			p("<TD> %s</TD></TR>" % (self._hdrs['Set-Cookie']))
		if self._hdrs.has_key('Set-Cookie2'):
			p("<TR><TH>Set-Cookie2 &nbsp;</TH>")
			p("<TD> %s</TD></TR>" % (self._hdrs['Set-Cookie2']))

		p("<TR><TH>Content-Length &nbsp;</TH><TD>")
		if self._hdrs.has_key('content-length'):
			p(" %s (%s)" % (print_k(self._hdrs.content_length), 
			  self._hdrs.content_length))
		else:
			p("&nbsp;" + self.no_data)
		if self._hdrs.get('content-length', -1) != self.size:
			p(" (<SPAN class='red'>actual size: %s</SPAN>)" % (self.size))
		p("</TD></TR>")

		p("<TR><TH>Server &nbsp;</TH><TD>")
		p(repr(self._hdrs.get('server', "&nbsp;" + self.no_data)))
		p("</TD></TR></TABLE>")

		p("<P class='desc'>%s</P>" % (description))

		if len(self._images):
			p('<H3>Images</H3>\n<UL>')
			for image in self._images:
				p(repr(image))
			p('</UL>')

		if len(self._objects):
			p('<H3>Objects</H3>\n<UL>')
			for object in self._objects:
				p(repr(object))
			p('</UL>')

		if len(self._frames):
			p('<H3>Frames</H3>')
			for frame in self._frames:
				p("<UL class='frame'>" + repr(frame) + "</UL>")
		p("</LI>")

		return join(o, "\n")


	def fetch_message(self, action):
		return '''
<script language="JavaScript">
<!--
window.status="%s %s";
// -->
</script>
''' % (action, self.request_url)


	def write_description(self):
		''' examines self and makes a rambling speech about its merits. ''' 
		o = []		# an array to hold sentences
		p = o.append
		
		if self.method not in ['GET', 'HEAD']:
			p("The %s method isn't cacheable. Change it to GET if possible." % (self.method))
			self.niceness = -10
		else:
			### redirect
			if self._redirectlevel:
				p("Clients have to follow %s redirect(s) to get to this object. This may be uncacheable, and causes slowdowns." % (self._redirectlevel))
				self.niceness = self.niceness - 1

			### freshness
			if self._hdrs.has_key('expires') or \
			  self._hdrs.get('cache-control', {}).has_key('max-age') or \
			  self._hdrs.get('cache-control', {}).has_key('s-maxage'):
				if self.freshtime <= 0:
					p("This object has been deliberately marked stale.")
					self.niceness = self.niceness -1
				else:
					p("This object will be fresh for %s." % (relative_date(self.freshtime, 0, 0)))
					self.niceness = self.niceness +1
			elif self._hdrs.get('last-modified', None) != None:
				p("This object doesn't have any explicit freshness information set, so a cache may use Last-Modified to determine how fresh it is with an adaptive TTL")
				f = []
				factor_list = self.lm_factors.keys()
				factor_list.sort()
				for factor in factor_list:
					ttl = relative_date(self.lm_factors[factor], 0, 0)
					f.append("%s (%i%%)" % (ttl, factor * 100))
				p("(at this time, it could be, depending on the adaptive percent used, considered fresh for: " + join(f, ", ") + ").")
			else:
				p("This object will be considered stale, because it doesn't have any freshness information assigned.")	
				self.niceness = self.niceness - 1

			### validation
			if self._hdrs.has_key('last-modified') or \
			  self._hdrs.has_key('etag'):
				if self._validated == 1:
					p('It can be validated with Last-Modified.')
					self.niceness = self.niceness + 1
				elif self._validated == 0:
					p("It has a validator present, but when a conditional request was made with it, the same object was sent anyway.")
					self.niceness = self.niceness - 1
				else:
					p("The object had changed when validation was attempted.")
			else:
				p("It doesn't have a validator present.")
				self.niceness = self.niceness - 1

			### cache-control
			if self._hdrs.get('cache-control', {}).has_key('public'):
				if self.authenticated:
					p("It's authenticated, but the Cache-Control: public header means that it will be cached.")
			elif self.authenticated:
				p("It's authenticated, so it won't be cached.")
				self.niceness = self.niceness - 1
			if self._hdrs.get('cache-control', {}).has_key('private'):
				p("It won't be cached by proxies, because it has a Cache-Control: private header.")
				self.niceness = self.niceness - 2
			if self._hdrs.get('cache-control', {}).has_key('no-cache'):
				p("It will be revalidated on every hit, because it has a Cache-Control: no-cache header.")
				self.niceness = self.niceness - 1
			if self._hdrs.get('cache-control', {}).has_key('no-store'):
				p("It won't be cached at all, because it has a Cache-Control: no-store header.")
				self.niceness = -10
			if self._hdrs.get('cache-control', {}).has_key('must-revalidate'):
				p("Because of the must-revalidate header, all caches will strictly adhere to any freshness information you set.")
			if self._hdrs.get('cache-control', {}).has_key('proxy-revalidate'):	
				p("Because of the proxy-revalidate header, proxy caches will strictly adhere to any freshness information you set.")

			### cookies
			if self._hdrs.has_key('Set-Cookie') or self._hdrs.has_key('Set-Cookie2'):
				p("This object requests that a Cookie be set; this makes it and other pages affected automatically stale; clients must check them upon every request.")
				self.niceness = -10

			### special cases
			if '?' in self.request_url:
				p("Because it contains a query ('?') in the URL, many caches will not keep this object.")
				self.niceness = self.niceness - 1
			if find(self.request_url, 'cgi-bin') != -1:
				p("Because it has 'cgi-bin' in the URL, many caches will not keep this object.")
				self.niceness = self.niceness - 1
				
			### content length
			if self._hdrs.has_key('content-length'):
				if self.size != self._hdrs.content_length:
					p("Its Content-Length header doesn't reflect the actual size of the object; this makes it difficult to work with.")
					self.niceness = self.niceness - 1
			else:
				p("It doesn't have a Content-Length header present, so it can't be used in a HTTP/1.0 persistent connection.")
				self.niceness = self.niceness - 1

			### clock skew
			if self.date_skew == None:
				if self._hdrs.has_key('date'):
					p("The Web server didn't send a correctly formed Date header for this object, which may cause problems.")
				elif not self.status in no_date:
					p("The Web server didn't send a Date header for this object, which may cause problems.")
			elif abs(self.date_skew) > self.skew_tolerance * 60:
				if self._hdrs.has_key('expires') or self._hdrs.has_key('last-modified'):
					p("The clock on this Web server appears to be set incorrectly; this can cause problems when calculating freshness.")

		return join(o, ' ')


	def link_list(self):
		''' return a HTML list of links '''
		if self.descend_frames: 
			descend = '&descend=on'
		else:
			descend = ''
		my_url = os.environ.get('SCRIPT_NAME', '')
		o = []
		tmp = {} 
		for link in self._links:
			link = urljoin(self.base_url, strip(link))
			(scheme, host, path, param, query, frag) = urlparse(link)
			if not scheme in ['http']:			
				continue
			tmp[link] = 1
		links = tmp.keys()
		links.sort()	# maybe this will be smarter one day...
		for link in links:
			o.append('<LI><A href="%s?query=%s%s">%s</A></LI>' % \
			  (my_url, quote(link), descend, link))
		for frame in self._frames:
			o = o + (frame.link_list())
		return o


	def date_skew_hash(self):
		''' Return a hash of servers with lists of their clock skews. '''
		if not self.status in no_date:
			skew_hash = { self.request_host: [self.date_skew] }
		for item in (self._images + self._objects):
			if item._errmsg: continue			# ignore error fetches
			if item.status in no_date: continue # ignore dateless status'
			try:
				skew_hash[item.request_host].append(item.date_skew)
			except KeyError:
				skew_hash[item.request_host] = [item.date_skew]
		for frame in self._frames:
 			for server, skewlist in frame.date_skew_hash().items():
				skew_hash[server] = skew_hash.get(server, []) + skewlist
		return skew_hash


	def date_skew_lists(self):
		'''
		End user of date_skew_hash(); returns a list servers that go over
		self.skew_tolerance (with their skews), and a list of servers that
		didn't give a date: header when they needed to.
		'''
		
		skew_servers = []
		no_date_servers = []
		for server, skew_list in self.date_skew_hash().items():
			if not len(skew_list): continue
			if None in skew_list:
				no_date_servers.append(server)
			abs_list = map(myabs, skew_list)
			max_skew = skew_list[abs_list.index(max(abs_list))]
			if myabs(max_skew) < self.skew_tolerance * 60: continue
			skew_servers.append((server, max_skew))
		return skew_servers, no_date_servers


def make_date(utime):
	''' take a unix epoch time and return a HTTP (RFC 1123) date. '''
	
	if not utime: return "None"
	return time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime(utime))


def relative_date(utime, now=None, show_sign=1):
	''' 
	Given two times, return a string that explains how far apart they are.
	show_sign can be:
	  0 - don't show
	  1 - ago / from now  [DEFAULT]
	  2 - early / late
	 '''

	signs = {	0:	('now', '', ''),
				1:	('now', 'ago', 'from now'),
				2:	('none', 'behind', 'ahead'),
			}

	if  utime == None: return None
	if now == None:	now = time.time()
	age = now - utime
	if age == 0: 
		return signs[show_sign][0]

	a = abs(age)
	wks = int(a / 60 / 60 / 24 / 7)
	day = int(a / 60 / 60 / 24) % 7
	hrs = int(a / 60 / 60) % 24
	min = int(a / 60) % 60
	sec = int(a % 60)

	if age > 0: sign = signs[show_sign][1]
	else: sign = signs[show_sign][2]
	if not sign: sign = signs[show_sign][0]

	arr = []
	if wks == 1: arr.append(str(wks) + ' week')
	elif wks > 1: arr.append(str(wks) + ' weeks')
	if day == 1: arr.append(str(day) + ' day')
	elif day > 1: arr.append(str(day) + ' days')
	if hrs: arr.append(str(hrs) + ' hr') 
	if min: arr.append(str(min) + ' min')
	if sec: arr.append(str(sec) + ' sec')
	arr = arr[:2]		# resolution
	if show_sign: arr.append(sign)
	return join(arr, ' ')


def print_date(utime, now=None):
	''' PrettyPrint dates (just for convenience)'''
	if utime:
		return make_date(utime), relative_date(utime, now)
	else:
		return "None", ''


def print_k(b):
	''' PrettyPrint bytes '''
	return "%1.1f" % (b / 1024.0) + 'K'

	
def word_wrap(paragraph, wrap=76):
	''' wrap a string on word boundries. '''
	from string import rindex, index
	last = len(paragraph)
	lines = []
	i = 0
	while 1:
		if i+wrap > last:
			lines.append(paragraph[i:last])
			break
		try:
			wrap_pt = rindex(paragraph, ' ', i, i+wrap)
		except ValueError:
			wrap_pt = index(paragraph, ' ', i)
		lines.append(paragraph[i:wrap_pt])
		i = wrap_pt + 1
	return join(lines, "\n")


def myabs(val):
	try:
		return abs(val)
	except TypeError:
		return 0


def html_header(title=''):
	''' standard HTML header '''
	return '''\
<HTML>
<HEAD>
<STYLE>
<!--
A:link, A:visited, A:active {
    color: #0000CC;
    text-decoration: none;
}

A:hover {
    text-decoration: underline;
}

BODY {
    margin: 6pt 72pt 36pt 3pt;
    padding: 0pt 0pt 0pt 0pt;
    color: black;
    background-color: white;
}

P {
    font-family: sans-serif;
    margin: 0pt 2pt 0pt 6pt;

}

P.desc, FORM {
    color: black;
    font-family: sans-serif;
    margin: 3pt 3pt 12pt 3pt;
}


TABLE {
    margin: 2pt 2pt 2pt 2pt;
}

TD {
    font-family: sans-serif;
    color: black;
    font-size: 0.9em;
    margin: 0pt 6pt 0pt 6pt;
	padding: 1pt 6pt 1pt 3pt;
	background-color: white;
}

TH {
    font-family: sans-serif;
    font-size: 0.9em;
    font-weight: normal;
    text-align: right;
    color: black;
    margin: 1px 6pt 1px 3pt;
    border-color: white;
    border-width: 1px 4px 1px 2px;
    border-style: solid;
    background-color: #DFDFDF;
}

H1, H2, H3, H4 {
    font-family: Tahoma, Verdana, sans-serif;
}

H1 {
    margin: 0pt 0pt 0pt 0pt;
    background-color: black;
    color: white;
    text-align: right;
    font-size: 1.3em;
    padding: 2px 9px 2px 18px;
    border: none;
    width: 100%%;
}

H2 {
    margin: 6pt 3pt 0pt 3pt;
    font-size: 1.2em;
    font-weight: normal;
}               

H2.section {
	font-size: 1.6em;
	font-weight: lighter;
}

H3 {
    margin: 3pt 12pt 0pt 12pt;
    font-size: 1.2em;
    font-weight: normal;
    color: black;
} 

H4 {
	margin: 3pt 12pt 0pt 12pt;
	font-size: 1.1em;
	font-weight: normal;
	color: #AAAAAA;
}

UL {
    margin-left: 36pt;
    list-style-type: disc;
}
 
UL.frame {
	color: #FFFFFF;
    border-color: #FFFFFF;
    border: 2px;
    border-style: solid;
    padding: 9px 18px 9px 24px;
    background-color: #CCCCCC;
}

UL.warninglist {
    margin-left: 9pt;
    margin-top: 2pt;
    margin-bottom: 9pt;
}

UL.list {
	font-family: sans-serif;
	margin-left: 3pt;
	font-size: 0.9em;
}

LI {
    font-family: sans-serif;
}

.green {
    color: #00AA00;
}

.yellow {
    color: #AAAA00;
}

.red {
    color: red;
}

.hide {
    display: none;
}

.warning {
    border-color: #FFFFFF;
    border: 2px;
    border-style: solid;
    margin-left: 8px;
    margin-top: 8px;
    margin-bottom: 24px;
    padding: 9px 9px 9px 9px;
    font-family: sans-serif;
    background-color: #FFD0D0;
}

.note {
    color: #999999;
    text-align: right;
    font-size: 0.8em;
}


EM {
    font-family: sans-serif;
    font-style: italic;
    font-weight: bold;
}

EM.bad {
    color: red;
}

EM.ok {
    color: green;
}

// -->
</STYLE>
<TITLE>%s</TITLE>
</HEAD>
<BODY>

<H1>%s</H1>

<P class="hide"><STRONG class="hide">
This service makes extensive use of Cascading Style Sheets.
Please enable them or use a CSS-aware browser to get the most benefit from it.
</STRONG></P>
''' % (title, title)
	
def html_footer():
	''' standardised HTML footer '''
	o = '''\

<P>
Need to know how to use this? Read 
<A href="http://www.mnot.net/cacheability/" target="_blank">about this service</A>, and
<A href="http://www.mnot.net/cache_docs/" target="_blank">how to make your site work with Web caches.</A>
</P>
'''
	if script_URL():
		o = o + '''\
<P>
You can also use this service as an Internet Explorer extension or a Netscape bookmark - 
<A href="%s?action=show_ext">find out how</a>.
</P>
''' % (script_URL())
	o = o + '''\
<P class="note">
version %s
</P>
<script language="JavaScript">
<!--
window.status="done.";
// -->
</script>
</BODY>
</HTML>
''' % (__version__)
	return o


def show_form():
	''' show the CGI query form '''
	return '''\
<FORM>
<P class="desc">This service allows you to check Web pages to see how caches
will handle them. </P>
<P class="desc">Enter the full URL of a Web page you wish to check.</P>
<INPUT name="query" size=40 value="http://">
<INPUT type="submit" value="check cacheability"><BR>
<INPUT type="checkbox" name="descend" checked> include referenced frames, images and objects<BR>
</FORM>
'''


def show_ext():
	''' show a page explaining how to install the extension '''
	return '''\

<TABLE width=90%% cellspacing=0 cellpadding=10><TR>
<td width=50%% valign=top>

<H2 class="section">Internet Explorer 4 and above</H2>
<P class="desc"> The Cacheability Engine can also be used as
an extension in the context-aware (right-click) menu for Internet Explorer 4
and above on Windows. This makes it easy to check cacheability as you surf.
To use it, follow these instructions.</P>

<H2>Step 1: Download</H2>
<P>Download the extension from <A href="%s?action=dl_ie_ext">here</a>.
The browser will prompt you to save a file 'cache.reg'; save it somewhere
convenient (like your desktop).
</P>

<H2>Step 2: Merge</H2>
<P>Now, right-click on the 'cache.reg' file. If there's a menu item called 'merge',
use it. If there isn't, you'll have to open a Windows Explorer and do it from
there.
</P>

<H2>Step 3: restart IE</H2>
<P>Finally, quit IE and restart it. When you find a page where you'd like to 
check the cacheability, right-click anywhere and select 'Cacheability'.
</P>

<H2>How does it work?</H2>
<P class="desc">This extension works because IE allows extesions to be added simply. It 
knows what page you're currently looking at, and submits it to the Internet
instance of the Cacheability Engine that you're using right now.</P>
</TD>

<TD width=50%% valign=top>
<H2 class="section">Netscape Navigator 4 and above</H2>
<P class="desc"> Netscape browsers can access the Engine by using a JavaScript
bookmark. When you wish to examine the cachability of a site, select the
bookmark (from the bookmarks menu, or the Personal Toolbar), and they
will appear in a new window.</P>
<P class="desc">To use it in this manner, follow these instructions.</P>

<H2>Step 1: Decide how to Save</H2> 
<P>Add the link below to your bookmark folder by <STRONG>one</STRONG> of the
following methods:</P>

<UL style="margin-left: 3pt">
<LI>Right-click on it and select "Add Bookmark"</LI>
<LI>Drag-and-drop the link <STRONG>into</STRONG> your bookmark menu.</LI>
<LI>Drag-and-drop the link onto your Personal Toolbar.</LI>
</UL>

<P>Once the bookmark is there, you can use this Engine by clicking on the 
saved bookmark when you have the target page in your front browser window.</P>


<H2>Save the Bookmark</H2>
<P>Here it is: 
<A HREF="javascript:MeRef=escape(location.href);Ca=open('','Ca','width=700,height=500,scrollbars,resizable,status');with(Ca){location.href='%s?version=%s&descend=on&query='+MeRef;void(x=1)}">Cacheability</A>
</P>


<H2>How does it work?</H2>

<P>The bookmark has JavaScript embedded in it that will automatically figure
out what page is in the current browser window, and access the Engine for
that page.</P>

<P>You can find more useful tools like this at 
<A HREF="http://www.bookmarklets.com/">http://www.bookmarklets.com/</a>.</P>


</TD>
</TR></TABLE>

<P><STRONG>Extension Problems?</STRONG> Try reinstalling as above - if the
Engine has been upgraded, you may need to change your extension.</P> <BR>


''' % (script_URL(), script_URL(), __version__)


def dl_ie_ext():
	''' download the ie extention '''
	return '''\
REGEDIT4

[HKEY_CURRENT_USER\Software\Microsoft\Internet Explorer\MenuExt]

[HKEY_CURRENT_USER\Software\Microsoft\Internet Explorer\MenuExt\&Cacheability]
@="%s?action=handle_ie_ext&version=%s"
''' % (script_URL(), __version__)

	
def handle_ie_ext():
	''' handle requests from the ie extension '''
	return '''\
<SCRIPT>
var win = external.menuArguments
var newWin = open("%s?version=%s&descend=on&query=" + escape(win.location) )
</SCRIPT>
''' % (script_URL(), __version__)


def http_headers(extra={}, content_type='text/html'):
	''' Generate HTTP headers '''
	h = ['Content-Type: %s' % (content_type)]
	for header, value in extra.items():
		h.append('%s: %s' % (header, value))
	h.append('')
	return join(h, "\n")


def script_URL():
	''' return the URL of the script from the environment '''
	env = os.environ
	try:
		return urlunparse(('http', env['SERVER_NAME'], env['SCRIPT_NAME'], 
		  None, None, None))
	except KeyError:
		return None
	

def cgi_main():
	''' run as a CGI script ''' 

	import cgi
	form = cgi.FieldStorage()
	if form.has_key('action'):
		action = form['action'].value
		if action == 'show_ext':
			print http_headers({'Cache-Control': 'max-age=86400'})
			print html_header('Using the Cacheability Engine as a Browser Extension')
			print show_ext()
			print html_footer()
			sys.exit()
		elif action == 'dl_ie_ext':
			print http_headers({'Content-Disposition': 'attachment; filename="cache.reg"'}, 
			  content_type='application/binary')
			print dl_ie_ext()
			sys.exit()
		elif action == 'handle_ie_ext':
			print http_headers({'Cache-Control': 'max-age=86400'})
			print handle_ie_ext()
			sys.exit()
	elif form.has_key('query'):
		query = form['query'].value
		print http_headers({'Cache-Control': 'max-age=30'})
		print html_header('Cacheability: %s' % (query))
		print '<P class="desc">The color of the list dot indicates the cacheability of the resource.</P>'
		sys.stdout.flush()
		f = HTMLSpider(query)
		if not form.has_key('descend'):
			f.descend_frames = 0
			f.run_images = 0
			f.run_objects = 0
		try:
			f.fetch()
			skew_servers, no_date_servers = f.date_skew_lists()
			if len(skew_servers):
				print "<DIV class='warning'>"
				if len(skew_servers) == 1:
					print "<STRONG>The following server's clock appears to be inaccurate.</STRONG> This can cause difficulties in caching its content. Please ask the administrator to synchronise its clock (this may also be caused by a proxy between you and the Web server)."
				else:
					print "<STRONG>These servers' clocks appear to be inaccurate.</STRONG> This can cause difficulties in caching their content. Please ask the administrator to synchronise their clocks (this may also be caused by a proxy between you and the Web server)."
				print "<UL class='warninglist'>"
				for (server, skew) in skew_servers:
					print "<LI><STRONG>%s</STRONG> - %s</LI>" % (server, relative_date(skew, 0, 2))
				print "</UL>"
				print "</DIV>"
			if len(no_date_servers):
				print "<DIV class='warning'>"
				if len(no_date_servers) == 1:
					print "<P><STRONG>The following server didn't issue a Date: header in the correct format (possibly not at all) for at least one request.</STRONG> Please ask the suppliers of your Web server software to fix this problem.</P>"
				else:
					print "<P><STRONG>These servers didn't issue a Date: header in the correct format (possibly not at all) for at least one request.</STRONG> Please ask the suppliers of your Web server software to fix this problem.</P>"
				print "<UL class='warninglist'>"
				for server in no_date_servers:
					print "<LI><STRONG>%s</STRONG></LI>" % (server)
				print "</UL>"
				print "</DIV>"
			print "<UL>", f, "</UL>"
		except Error, errmsg:
			if type(errmsg) is types.InstanceType:
				try:
					errmsg = errmsg.args[-1]
				except:
					pass
			elif type(errmsg) is type(()):
				errmsg = errmsg[-1]
			print "<H3>Error: %s.</H3>" % (errmsg)
		print '<HR width=68% align=left noshade color=black size=1>'
		link_list = f.link_list()
		if len(link_list):
			print "<H4>Check cacheability for:</H4>"
			print '<UL class="list">'
			print join(link_list, "\n")
			print '</UL>'
		print '<FORM><INPUT TYPE="submit" value=" Check Another URL "></FORM>\n'
		print html_footer()
	else:
		print http_headers({'Cache-Control': 'max-age=86400'})
		print html_header('Cacheability Query')
		check_version()
		print show_form()
		print html_footer()


def cgi_wrap():
	''' wraps cgi_main in a safe, error-reporting env '''
	sys.stderr = sys.stdout
	try:
		cgi_main()
	except SystemExit:
		pass
	except:
		print '''	
<BR clear=all> <H2>There has been a unrecoverable error.</H2> <P>It probably
wasn't your fault, but the following may give you an idea of what was
happening when things went bad. Please mail this page to 
<A HREF="mailto:errors@mnot.net">errors@mnot.net</a></P>
<P>(version %s)</P> 
<PRE> 
''' % (__version__)

		import traceback
		traceback.print_exc()


def clui_main():
	''' run as a command line utility '''

	try:
		check_version()
	except StandardError, errmsg:
		sys.stderr.write(str(errmsg) + "\n")
		sys.exit()

	import getopt
	opthash = {}
	try:
		(optlist, args) = getopt.getopt(sys.argv[1:], 'vrh')
	except getopt.error:
		clui_usage()
	if not len(args):
		clui_usage()
	for pair in optlist:
		opthash[pair[0]] = pair[1]	
	### help?
	if opthash.has_key('-h'):
		clui_usage()
	### verbose?
	verbose = 0
	if opthash.has_key('-v'):
		verbose = 1
	### recurse?
	recurse = 0
	if opthash.has_key('-r'):
		recurse = 1

	for query in args:
		try:
			f = Spider(query)
			if not verbose:
				f.verbose = 0
			if not recurse:
				f.descend_frames = 0
				f.run_images = 0
				f.run_objects = 0
			f.fetch()
			print
			print f
		except Error, errmsg: 
			if type(errmsg) is types.InstanceType:
				errmsg = errmsg.args[-1]
			print errmsg
			

def clui_usage():
	print '''\

Usage: %s [ -v ] [ -r ] url [ url... ]
Options:
	-v  verbose
	-r  recurse frames, images and objects
	-h  this help message

Version %s

''' % (sys.argv[0], __version__)
	sys.exit()

def check_version():
	try:
		if sys.hexversion < 17105648:
			raise StandardError 
	except:
		raise StandardError, 'The Engine requires Python 1.5.2 or greater. Please upgrade.'


if __name__ == '__main__':
	if os.environ.has_key('GATEWAY_INTERFACE'):
		cgi_wrap()
	else:
		clui_main()
