#!/usr/local/bin/python

# (c) 1999 Copyright Mark Nottingham
# <mnot@pobox.com>
#
# This software may be freely distributed, modified and used,
# provided that this copyright notice remain intact.
#
# This software is provided 'as is' without warranty of any kind.

'''

HTTP Header Parsing and Representation

The Headers() class wraps the HTTP Header Types, allowing easy and automated
access to them. It can be instantiated with an argument of either a
rfc822.Message instance or a string containing headers to be parsed. 

Once the headers have been parsed (or an empty instance has been created), 
each header can be accessed as an attribute. Dashes ('-') in the header's
name should be converted to underscores ('_'), and the names are not 
case-sensitive.

Depending on a particular header's type, it may be accessed as a string,
integer, list or dictionary; for full details, look at the header's doc
string.

If you need to create a header, it must first be instantiated within the
Headers() instance. You do this with an argument of the data type requested.
For instance:
  f = Headers()
  f.cache_control = {}
  f.cache_control.max_age = 600

Finally, the putheaders() method can be used to pass the headers back
to the specified httplib object, using its putheader() method.

LIMITATIONS
Note that although the header classes attempt to format their contents
correctly, they do not enforce HTTP compliance; the user should be familiar
with the specification (RFC 2616).

'''



import rfc822, sys
from string import lower, join, split, replace
from UserDict import UserDict
from httpheadertypes import *


__version__ = "0.30"


class Headers(UserDict):
	'''
	Contains a dictionary of Header instances, keyed by header name
	(case-insensitive).

	Can be instantised with a rfc822.Message object, or text of the headers,
	as the argument.
	'''

	def __init__(self, message=None):		
		UserDict.__init__(self)
		self.errparse = []
		self.errprint = []
		if message != None:
			if isinstance(message, rfc822.Message):
				header_list = message.headers
			else:
				header_list = split(message, "\n")
			for line in header_list:
				try:
					k, v = split(line, ":", 1)
					header_name, value = get_header_name(strip(k)), strip(v)
					if self.data.has_key(header_name) and \
					  isinstance(self.data[header_name], List):
						self.data[header_name].append(
						  self.data[header_name].objtype(value))
					else:
						self.data[header_name] = \
						  header_names.get(header_name, Str)(value)
				except IOError:
					self.handle_exception('parse', line)

	def __getitem__(self, key):
		return self.data[get_header_name(key)]

	def __setitem__(self, key, item):
		first_seen = 0
		header_name = get_header_name(key)
		if not self.data.has_key(header_name):
			first_seen = 1
			self.data[header_name] = header_names.get(header_name, Str)()
		if isinstance(item, GenericHeader):
			self.data[header_name] = item
		elif type(item) is type(self.data[header_name].data):
			self.data[header_name].data = item
		else:
			if first_seen:
				del self.data[header_name]
			raise TypeError, 'must be %s' % (
			  str(type(self.data[header_name].data)))

	def __delitem__(self, key):
		del self.data[get_header_name(key)]

	def __getattr__(self, key):
		return self[replace(key, '_', '-')]
		
	def __setattr__(self, key, item):
		if key in ['data', 'errparse', 'errprint']:
			self.__dict__[key] = item
		else:
			self[replace(key, '_', '-')] = item

	def __delattr__(self, key):
		del self[replace(key, '_', '-')]

	def has_key(self, key):
		return self.data.has_key(get_header_name(key))
	
	def get(self, key, default):
		return self.data.get(get_header_name(key), default)
		
	def __repr__(self):
		o = []
		for header, obj in self.data.items():
			try:
				o.append("%s: %s" % (header, repr(obj)))
			except Exception, args:
				self.handle_exception('repr', header)
		return join(o, "\n")

	def putheaders(self, httpobj):	
		''' put contained headers into httpobj using its putheader() meth. '''
		for header, obj in self.data.items():
			httpobj.putheader(header, repr(obj))

	def handle_exception(self, what, info):
		'''
		Handles parsing and repr errors (with bad headers or values). 
		Raise an exception to stop processing; to continue, log the error
		as you wish.
		'''

		exception = sys.exc_info()
		if what == 'parse': arr = self.errparse
		else: arr = self.errprint
		arr.append((what, info))
		exception[1].args = (exception[1].args, info)
		raise exception[1]



