#!/usr/local/bin/python

# (c) 1998 Copyright Mark Nottingham
# <mnot@pobox.com>
#
# This software may be freely distributed, modified and used,
# provided that this copyright notice remain intact.
#
# This software is provided 'as is' without warranty of any kind.


'''

HTTP Header Types

This module contains the GenericHeader class, which can be (and is) 
subclassed to define the behaviour of a HTTP header. The class which
handles a particular header's representation is determined by the
header_names dictionary.

Each of the classes has the same basic interface:

* a parse() method - which will parse the header payload fed to it (and
  raise errors as necessary)
* a data dictionary - which allows access to the raw data after parsing
* a __repr__ method  - to print the data correctly in the Header's format

If the object is instantiated with an argument, it is parsed automatically.

Many of the headers are Lists, which are comma-seperated. Because the List
class is a subclass of the UserList object, it can be accessed like a normal
list. The same holds true for the Dict class, with respect to dictionaries.

Each class has a short __doc__ string that describes any special attributes
or behaviours.

the get_header_name function returns the canonical capitalisation of the
header name fed to it, if available, or a capitalied version if not.


TODO
- QValList - sort / get_best
- Warning = list of (int str quostr quohttpdate) ; (currently List)
- Expect = list of token[=qs] [;param] OR 100-continue ; (currently List)
- digest authorization
- Set-Cookie only takes the last one. Stupid !@#%$ Netscape spec!
- Range is a pretty pitiful implementation so far.
- Content-Range  ; (currently Str)
'''

import time, re, rfc822, urlparse, base64
from string import lower, join, strip, split, replace, capitalize
from operator import add, sub, mul, div, neg
from UserList import UserList
from UserDict import UserDict
from cStringIO import StringIO


__version__ = "0.30"



### patterns for splitting 

commapat = re.compile(r"""
	(?:,\s*|^)
	(
	(?:
	"(?:\\"|[^"])*"	
	|
	[^,"]+
	)*
	)
""", re.VERBOSE)

def splitoncomma(instr):
	if not instr: return []
	return commapat.findall(instr)	

semipat = re.compile(r"""
	(?:;\s*|^)
	(
	(?:
	"(?:\\"|[^"])*"	
	|
	[^;"]+
	)*
	)
""", re.VERBOSE)

def splitonsemi(instr):
	if not instr: return []
	return semipat.findall(instr)





class GenericHeader:
	'''
Base class for HTTP Headers. Should be subclassed.
'''
	def __init__(self, instr=None):
		self.raw = instr
		if instr != None: self.parse(instr)
	def __setattr__(self, attr, val):
		if attr == 'data' and hasattr(self, 'data') and \
		  type(val) != type(self.data):
			try:		### let int and float intermingle
				int(val)
				float(val)
			except:
				if val != None:
				  	raise TypeError, "must be %s" % (str(type(self.data)))
		self.__dict__[attr] = val
	def __cmp__(self, other):
		return cmp(self.data, other)
	def __add__(self, other):
		return self._math(other, add)
	def __radd__(self, other):
		return self._rmath(other, add)
	def __sub__(self, other):
		return self._math(other, sub)		
	def __rsub__(self, other):
		return self._rmath(other, sub)
	def _math(self, other, meth):
		if isinstance(other, GenericHeader):
			tmp = self.__class__()
			tmp.data = meth(self.data, other.data)
			return tmp
		elif isinstance(other, type(other)):
			tmp = self.__class__()
			tmp.data = meth(self.data, other)
			return tmp
		else:
			raise TypeError
	def _rmath(self, other, meth):
		if isinstance(other, GenericHeader):
			tmp = self.__class__()
			tmp.data = meth(other.data, self.data)
			return tmp
		elif isinstance(other, type(other)):
			tmp = self.__class__()
			tmp.data = meth(other, self.data)
			return tmp
		else:
			raise TypeError
	def parse(self, instr):
		self.data = instr		
	def __repr__(self):
		return repr(self.data)
	def __nonzero__(self):
		if self.data: return 1
		return 0


class Int(GenericHeader):
	'''
Integer.
'''
	def __init__(self, instr=None):
		self.data = 0
		GenericHeader.__init__(self, instr)
	def __mul__(self, other):
		return self._math(other, mul)	
	def __rmul__(self, other):
		return self._rmath(other, mul)	
	def __div__(self, other):
		return self._math(other, div)
	def __rdiv__(self, other):
		return self._rmath(other, div)
	def __int__(self):
		return int(self.data)
	def __float__(self):
		return float(self.data)
	def __abs__(self):
		return abs(self.data)
	def __neg__(self):
		return neg(self.data)
	def parse(self, instr):
		self.data = int(instr)
	def __repr__(self):
		return repr(int(self.data))


class Str(GenericHeader):
	'''
String (case-sensative)
'''
	def __init__(self, instr=None):
		self.data = ''
		self.case_sens = 1
		GenericHeader.__init__(self, instr)	
	def parse(self, instr):
		if not self.case_sens:
			instr = lower(instr)
		self.data = str(instr)
	def __getitem__(self, i):
		return self.data[i]
	def __getslice__(self, i, j):
		return self.data[i:j]
	def __repr__(self):
		if self.case_sens:
			return str(self.data)
		else:
			return lower(str(self.data))


class Str_i(Str):
	'''
String (case-insensitive)
'''
	def __init__(self, instr=None):
		Str.__init__(self, instr)
		self.case_sens = 0


class List(GenericHeader, UserList):
	'''
List of case-sensitive strings.
'''
	def __init__(self, instr=None, objtype=Str):
		self._objtype = objtype
		self.data = UserList()
		GenericHeader.__init__(self, instr)
	def __setattr__(self, key, item):
		if key in ['data']:
			d = []
			for i in item:
				d.append(self._objtype())
				d[-1].data = i
			self.__dict__['data'] = d
		else:
			self.__dict__[key] = item		
	def append(self, item):
		if isinstance(item, self._objtype):
			self.__dict__['data'].append(item)
		else:
			self.__dict__['data'].append(self._objtype())
			try:
				self.__dict__['data'][-1].data = item
			except TypeError, why:
				del self.__dict__['data'][-1]
				raise TypeError, why
	def insert(self, i, item):
		self.__dict__['data'].insert(i, self._objtype())
		try:
			self.__dict__['data'][i].data = item		
		except TypeError, why:
			self.__dict__['data'][i]
			raise TypeError, why
	def parse(self, instr):
		for header in  map(strip, splitoncomma(instr)):
			self.append(self._objtype(header))
	def __repr__(self):
		return join(map(repr, self.data), ', ')


class List_i(List):
	'''
List of case-insensitive strings.
'''
	def __init__(self, instr=None):
		List.__init__(self, instr, Str_i)
	

class Dict(GenericHeader, UserDict):
	'''
Dictionary of key=value pairs, separated by commas. Keys are 
case-insensitive; values are not. If possible, the value will be
converted to an integer.
'''
	def __init__(self, instr=None, splitfunc=splitoncomma, joinchar=','):
		self.data = {}
		self._splitfunc = splitfunc
		self._joinchar = joinchar
		GenericHeader.__init__(self, instr)
	def __getattr__(self, attr):
		if attr[0] == '_':
			return self.__dict__[attr]
		attr = lower(replace(attr, '_', '-'))
		return self.data[attr]
	def __setattr__(self, attr, val):
		if attr[0] == '_' or attr == 'data':
			self.__dict__[attr] = val
		else:
			attr = lower(replace(attr, '_', '-'))
			self.__dict__['data'][attr] = val
	def __delattr__(self, attr):
		if attr[0] == '_':
			del self.data[attr]
		else:
			attr = lower(replace(attr, '_', '-'))
			del self.data[attr]
	def parse(self, instr):
		o = self._splitfunc(instr)
		res = {}
		for i in o:
			try:
				name, value = split(strip(i), '=', 1)
				name = lower(name)
				value = unquotestring(value)
				### special cases for Cache-Control
				if name in ['private', 'no-cache']:
					value = map(get_header_name, self._splitfunc(value))
				try:
					value = int(value)
				except:
					pass
			except ValueError:
				name, value = lower(strip(i)), None
			res[name] = value
		self.data = res		
	def __repr__(self):
		o = []
		for k, v in self.data.items():
			k = lower(k)
			if v == None:
				o.append(k)
			else:
				if k in ['private', 'no-cache'] and type(v) == type([]):
					v = '"%s"' % (
					  join(map(get_header_name, v), self._joinchar + " "))
				elif type(v) is type(''):
					v = quotestring(v)
				o.append("%s=%s" % (k,v))
		return join(o, self._joinchar + " ")


class SemiDict(Dict):
	def __init__(self, instr=None):
		Dict.__init__(self, instr, splitonsemi, ';')


class SemiDictList(List):
	def __init__(self, instr=None):
		List.__init__(self, instr, SemiDict)


class ETag(Str):
	'''
ETag. The value set will be quoted (and escaped, if need be). The .weak
attribute denotes a weak validator.
'''
	def __init__(self, instr=None):
		self.weak = 0
		Str.__init__(self, instr)
	def parse(self, instr):
		if instr[:2] == 'W/':
			self.weak = 1
			instr = instr[2:]
		self.data = unquotestring(instr)
	def __repr__(self):
		weak = ''
		if self.weak:
			weak = "W/"
		return "%s%s" % (weak, quotestring(self.data, force=1))


class ETagList(List):
	'''
List of ETags.
'''
	def __init__(self, instr=None):
		List.__init__(self, instr, ETag)


class HttpDate(Int):
	'''
HTTP Date - available as seconds since Epoch.
'''
	def parse(self, instr):
		self.data = None	# we start with None to signify an invalid date
		parsed = rfc822.parsedate(instr)
		if type(parsed) == type(()):
			try:
				self.data = timegm(parsed)
			except:
				pass
	def __repr__(self):
		try:
			return time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime(self.data))		
		except:
			return "invalid date (%s)" % self.data

class Uri(GenericHeader):
	'''
URI - available as urlparse 6-tuple.
'''
	def __init__(self, instr=None):
		self.data = ('', '', '', '', '', '')
		GenericHeader.__init__(self, instr)
	def parse(self, instr):
		self.data = urlparse.urlparse(instr)
	def __repr__(self):
		return urlparse.urlunparse(self.data)



class Param(Str_i):
	'''
Parameterized data. primary data available as object instance, 
parameters available as a dictionary in the .params attribute.
Param keys are case-insensitive, values are case-sensitive.
'''
	def __init__(self, instr=None):
		self.params = {}
		Str_i.__init__(self, instr)
	def parse(self, instr):
		try:
			self.data, args = split(instr, ';', 1)
			self.data = lower(self.data)
		except ValueError:
			self.data = lower(instr)
			return
		for param in splitonsemi(args):
			try:
				attr, value = split(param, "=", 1)
				self.params[lower(attr)] = unquotestring(value)
			except ValueError:
				self.params[lower(param)] = None
	def __repr__(self):
		o, out = [], ''
		for k,v in self.params.items():
			if v == None:
				o.append(k)
			else:
				o.append("%s=%s" % (k, quotestring(v)))
		if o:
			out = '; ' + join(o, '; ')
		return "%s%s" % (lower(self.data), out)


class ParamList(List):
	def __init__(self, instr=None):
		List.__init__(self, instr, Param)


class QValList(ParamList):
	'''
List of parameterised data. If an item has a 'q' parameter associated with 
it, that will determine its relative value.

The sort() method can be used to sort the list by qvalues, highest last.
The get_best() method can be used to determine which candidate in a list
fed to it is most acceptable, based on qvalues.
'''
	def __init__(self, instr=None):
  		ParamList.__init__(self, instr)
	def sort(self):
		pass
	def get_best(self, candidates):
		pass


class Header(Str):
	'''
Canonicalized HTTP header name.
'''
	def parse(self, instr):
		self.data = get_header_name(instr)
	def __repr__(self, instr):
		return get_header_name(self.data)


class HeaderList(List):                ### list of headers, or *
	'''
List of HTTP headers in canonical format.
'''
	def __init__(self, instr=None):
		List.__init__(self, instr, Header)


class Challenge(Dict):
	'''
HTTP Authorization challenge. Currently, only Basic is supported.
Attributes:
  .scheme : scheme used (should be set to 'basic')
  .realm : authentication realm
'''
	def __init__(self, instr=None):
		Dict.__init__(self, instr)
	def parse(self, instr):
		try:
			scheme, args = split(instr, None, 1)
		except:
			scheme, args = instr, None			
		self.data['scheme'] = lower(scheme)
		Dict.parse(self, args)
	def __repr__(self):
		if lower(self.data['scheme']) == 'basic':
			o = []
			for k,v in self.data.items():
				if lower(k) == 'scheme': continue
				o.append("%s=%s" % (k, quotestring(v)))
			return "%s %s" % (
			  capitalize(lower(self.data['scheme'])), join(o, ', '))
		else:
			raise StandardError, "Digest Auth not supported yet."			  


class Credentials(Dict):
	'''
HTTP Authorization credentials. Currently, only Basic is supported.
Attributes:
  .scheme : scheme used (should be set to 'basic')
  .username : username
  .password : password
'''
	def __init__(self, instr=None):
		Dict.__init__(self, instr)
	def parse(self, instr):
		scheme, args = split(instr, None, 1)
		self.data['scheme'] = lower(scheme)
		if self.data['scheme'] == 'basic':
			self.data['username'], self.data['password'] = split(
			  base64.decodestring(args), ':', 1)
		else:
			raise StandardError, "Digest Auth not supported yet."
	def __repr__(self):
		if lower(self.data['scheme']) == 'basic':
			return "%s %s" % (
			  capitalize(lower(self.data['scheme'])), 
			  base64.encodestring(join((self.data['username'], 
			  self.data['password']), ':'))[:-1])
		else:
			raise StandardError, "Digest Auth not supported yet."



class Range(List):
	'''
Range request. Specifier is available on .specifier attribute (should be
'bytes'); ranges are in the List itself.
'''
	def __init__(self, instr=None):
		self.specifier = 'bytes'
		List.__init__(self, instr, Str)
	def parse(self, instr):
		specifier, args = split(instr, '=', 1)
		self.specifier = lower(specifier)
		List.parse(self, args)
	def __repr__(self):
		return "%s=%s" % (self.specifier, List.__repr__(self))


class HttpDateorInt(HttpDate, Int):
	'''
Either a HTTP Date or Integer. If a date is represented as seconds since
the epoch, the .is_date attribute should be true; if it represents delta
seconds, it should be false.
'''
	def __init__(self, instr=None):
		self.is_date = 0
		HttpDate.__init__(self, instr)
	def parse(self, instr):
		try:
			Int.parse(self, instr)
		except:
			HttpDate.parse(self, instr)
			self.is_date = 1
	def __repr__(self):
		if self.is_date:
			return HttpDate.__repr__(self)
		else:
			return Int.__repr__(self)


class Validator(HttpDate, ETag):
	'''
Either a HTTP Date or an ETag. If a date is represented, the .is_date
attribute should be true; otherwise, it should be false.
'''
	def __init__(self, instr=None):
		self.is_date = 0
		self.weak = 0
		HttpDate.__init__(self, instr)
	def parse(self, instr):
		try:
			HttpDate.parse(self, instr)
			self.is_date = 1
		except:
			self.__dict__['data'] = ''
			ETag.parse(self, instr)
	def __repr__(self):
		if self.is_date:
			return HttpDate.__repr__(self)
		else:
			return ETag.__repr__(self)




### header_names: this is where the headers are actually associated with
### an HeaderType, and shown with their canonical spelling.


header_names = {'Accept': QValList,
				'Accept-Charset': QValList,
				'Accept-Encoding': QValList,
				'Accept-Language': QValList,
				'Accept-Ranges': List_i,
				'Age': Int,
				'Allow': List,
				'Authorization': Credentials,
				'Cache-Control': Dict,
				'Connection': List_i,
				'Content-Encoding': List_i,
				'Content-Language': List_i,
				'Content-Length': Int,
				'Content-Location': Uri,
				'Content-MD5': Str,
				'Content-Range': Str,
				'Content-Type': Param,
				'Cookie': SemiDictList,
				'Date': HttpDate,
				'ETag': ETag,
				'Expect': List,
				'Expires': HttpDate,
				'From': Str,
				'Host': Str_i,
				'If-Match': ETagList,
				'If-Modified-Since': HttpDate,
				'If-None-Match': ETagList,
				'If-Range': Validator,
				'If-Unmodified-Since': HttpDate,
				'Last-Modified': HttpDate,
				'Location': Uri,
				'Max-Forwards': Int,
				'Pragma': Dict,
				'Proxy-Authenticate': Challenge,
				'Proxy-Authorization': Credentials,
				'Range': Range,
				'Referer': Uri,
				'Retry-After': HttpDateorInt,
				'Server': Str,
				'Set-Cookie': SemiDict,
				'Set-Cookie2': SemiDictList,
				'TE': QValList,
				'Trailer': HeaderList,
				'Transfer-Encoding': List_i,
				'Upgrade': List,
				'User-Agent': Str,
				'Vary': HeaderList,
				'Via': List,
				'Warning': List,
				'WWW-Authenticate': Challenge,
				}



def get_header_name(query):
	'''
	Given a header in any case, return its properly capitalised version.
	'''
	try:
		return header_names.keys() \
		  [map(lower, header_names.keys()).index(lower(query))]
	except (KeyError, ValueError):
		return capitalize(lower(query))




### Utility Functions


def timegm(tmtuple):
	''' returns epoch seconds from a GMT time tuple. '''
	import calendar
	EPOCH = 1970
	year, month, day, hour, minute, second = tmtuple[:6]
	if year < EPOCH:
		if year < 69:
			year = year + 2000
		else:
			year = year + 1900
		if year < EPOCH:
			raise ValueError, 'invalid year'
	if not 1 <= month <= 12:
		raise TypeError, 'invalid month'
	days = 365 * (year-EPOCH) + calendar.leapdays(EPOCH, year)
	for i in range(1, month):
		days = days + calendar.mdays[i]
	if month > 2 and calendar.isleap(year):
		days = days + 1
	days = days + day - 1
	hours = days * 24 + hour
	minutes = hours * 60 + minute
	seconds = minutes * 60 + second
	return seconds
		

def quotestring(instr, force=0):
	''' does NOT quote control characters. '''
	if not force and \
	  '"' not in instr and \
	  ',' not in instr and \
	  ';' not in instr:
			return instr
	if instr == '*':
			return instr
	return '"%s"' % (re.sub(r'"', r'\\"', instr))


def unquotestring(instr):
	''' does NOT unquote control characters. '''
	instr = strip(instr)
	if not instr or instr == '*':
		return instr
	if instr[0] == instr[-1] == '"':
		instr = instr[1:-1]
		instr = re.sub(r'\\"', r'"', instr)
	return instr


