
'''
Web Cacheability Engine - Spider Module
Copyright 1999 Mark Nottingham <mnot@pobox.com>

THIS SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND.

This software may be freely distributed, but may not be modified in any way.
See the LICENSE file for more information.
'''

# This file was edited with four space tab stops. No tabs were harmed during 
# the production of this program.

### imports
import htmllib, time, sys, os, types, socket, httplib, formatter
from string import lower, join, split, strip, atoi, maketrans, translate
from urlparse import urlparse, urlunparse, urljoin
from httpheaders import Headers

### determine if it's safe to use signal. I've tried playing with
### select(), but it breaks too unpredictably and messily on too
### many platforms, at least in Python.
signal = None
if os.name == 'posix':
	import signal


#############################################################################
### Configuration

timeout = 30			# timeout for HTTP requests (in seconds)
proxy = None			# specify a proxy cache

#############################################################################
### Hardcoded settings: do not change!

Error = 'Error'
__version__ = '1.30'
UserAgentString = 'CacheabilityEngine/%s <http://www.mnot.net/cacheability/>' % (__version__)

http_status_codes = {	400:	'Bad Request',
						401:	'Unauthorized (password needed)',
						403:	'Forbidden',
						404:	'Object Not Found',
						500:	'Server Error',
					}

### status codes which don't require a date header
no_date = [100,101,500,501,502,503,504,505]

### cdata translation string
cdata_tr = maketrans('\t\n', '  ')


#############################################################################


class Spider:
	''' 
	Class to contain a HTTP object. Main method of interest is fetch().
	By default, outputs wrapped text; subclass to override.
	'''

	def __init__(self, url):
		self.request_url = strip(url)	# full URL that we contain
		self.descend_frames = 3		# number of framesets to recurse, max
		self.resolve_redirects = 5	# number of redirects to resolve, max
		self.validate = 1			# do we want to validate where appropriate?
		self.run_images = 1			# do we want to get images as well?
		self.run_objects = 1		# do we want to get other content as well?
		self.verbose = 1			# should we print messages?
		self.skew_tolerance = 5		# how much skew before clock is bad (mins)
		self.no_data = '-'			# what to print when there's no value

		self.method = 'GET'			# hook for future methods
		self.status = None			# response code from the server
		self.authenticated = 0		# flag to say if the page is authenticated

		self.base_url = None		# base URI for referenced objects
		self.content_hash = None	# md5 hash of body
		self.title = ''				# page title
		self.date_skew = None		# clock skew (None if no Date hdr)
		self.size = None			# object body size (bytes)
		self.lm_age = None			# object age (seconds) NOT HTTP AGE
		self.freshtime = None		# TTL (seconds)
		self.niceness = -1			# flag for how cacheable the obj is 

		self._errmsg = None			# holds any error message
		self._validated = None		# holds validation sucess info
		self._redirectlevel = 0		# holds redirect resolution count
		self._body = None			# holds object body
		self._hdrs = None			# holds object headers
		self._images = []			# holds referenced image objects
		self._frames = []			# holds referenced html objects
		self._objects = []			# holds other objects referenced (.js, etc)
		self._meta = {}				# holds page's meta tags
		self._links = []			# holds page's links

		# set up lm_factors to apply for adaptive TTL
		self.lm_factors = {	.2: None,
							.5: None,
							 1: None,
						}


	def __copy_config(self, tmp):
		''' 
		Copies selected attributes into a nomintated 'child' object for 
		recursion. Not the prettiest way to do this, but it works. 
		'''

		tmp.descend_frames = self.descend_frames
		tmp.resolve_redirects = self.resolve_redirects
		tmp.validate = self.validate
		tmp.run_images = self.run_images
		tmp.run_objects = self.run_objects
		tmp.verbose = self.verbose


	def process_url(self):
		''' standardise the URL '''

		### normalise the url.
		(	self.request_scheme,
			self.request_host,
			self.request_path,
			self.request_parameters,
			self.request_query,
			self.request_fragment	) = urlparse(self.request_url)

		if self.request_scheme != 'http':
			raise Error, 'non-HTTP protocols not supported'

		self.request_host = lower(self.request_host)

		self.request = urlunparse((	'', '',
									self.request_path,
									self.request_parameters,
									self.request_query,
									self.request_fragment	))
		if not self.base_url:
			self.base_url = urlunparse(('http', self.request_host, 
			  self.request, '', '', ''))


	def fetch_message(self, action):
		''' message which is printed during each fetch, if verbose. '''
		return "%s %s" % (action, self.request_url)


	def fetch(self, IMS=None):
		''' fetch the page from the Web and do initial attr population '''

		import md5
		self.process_url()

		### if we want to, we can print out progress of how we're going
		if self.verbose:
			if not IMS:
				print self.fetch_message("fetching")
			else:
				print self.fetch_message("validating")
			sys.stdout.flush()

		if signal:
			signal.signal(signal.SIGALRM, handle_alarm)

		### and here we go
		try:
			if signal:
				signal.alarm(timeout)
			try:
				if proxy:
					h = httplib.HTTP(proxy)
				else:
					h = httplib.HTTP(self.request_host)
			except socket.error, errtext:
				if signal:
					signal.alarm(0)
				raise Error, errtext
			if proxy:
				h.putrequest('GET', self.request_url)
			else:
				h.putrequest('GET', self.request)
			h.putheader('Host', self.request_host)
			h.putheader('User-Agent', UserAgentString)
			h.putheader('Accept', '*/*')
			h.putheader('Pragma', 'no-cache')
			h.putheader('Cache-Control', 'no-cache')
			if IMS:
				h.putheader('If-Modified-Since', repr(IMS))
			h.putheader('Connection', 'close')
			h.endheaders()
			errcode, errmsg, i = h.getreply()
			if errcode == -1:
				errmsg = 'HTTP 0.9 server - please upgrade'

			### process headers
			self._hdrs = Headers(i)
			self.status = errcode
			self.date_fetched = self._hdrs.get('date', None) or time.time()

			### clock skew
			try:
				self.date_skew = int(self._hdrs.date - time.time())
			except (KeyError, TypeError):		# no or bad date header
				pass
	
			### calculate age (NOT HTTP Age - age of the object on the server)
			if self._hdrs.has_key('date') and \
			  self._hdrs.get('last-modified', None) != None:
				self.lm_age = self.date_fetched - \
				  self._hdrs['last-modified']

			### calculate explicit freshness
			if self._hdrs.has_key('expires'):
				if self._hdrs.expires == None:
					expires_freshtime = - self.date_fetched
				else:
					expires_freshtime = self._hdrs.expires - self.date_fetched
			else:
				expires_freshtime = None
			self.freshtime = \
			  self._hdrs.get('cache-control', {}).get('s-maxage', None) or \
			  self._hdrs.get('cache-control', {}).get('max-age', None) or \
			  expires_freshtime
			# 0 is a special case
			if \
			  self._hdrs.get('cache-control', {}).get('s-maxage', None) == 0 or \
			  self._hdrs.get('cache-control', {}).get('max-age', None) == 0:
				self.freshtime = 0

			### calculate adaptive freshness
			if self.lm_age:
				for factor in self.lm_factors.keys():
					self.lm_factors[factor] = (factor * self.lm_age)
	
			if signal:
				signal.alarm(0)
		except socket.error, why:
			raise Error, why

		if errcode == 200:								### good response
			try:
				self._body = h.getfile().read()
			except IOError, why:
				raise Error, why
			self.content_hash = md5.new(self._body).digest()
			self.size = len(self._body)
			if self.validate and self._hdrs.has_key('last-modified') and \
			  self._hdrs.last_modified != None:
				tmp = self.__class__(self.request_url)
				self.__copy_config(tmp)
				tmp.validate = 0
				val_response = tmp.fetch(self._hdrs.last_modified)
				if val_response == 304:
					self._validated = 1
				elif val_response == 200 and tmp.content_hash == self.content_hash:
					self._validated = 0
			if self._hdrs.get('content-type', None) == 'text/html' and not IMS:
				self.parse_html()
			return 200

		elif errcode == 304:							### not modified
			return 304

		elif errcode in [301, 302, 303, 307]:			### redirect
			if self.resolve_redirects:
				if self._redirectlevel <= self.resolve_redirects:
					try:
						location = repr(self._hdrs.location)
					except:
						raise Error, 'redirect with no Location'
					tmp = self.__class__(urljoin(self.request_url, location))
					self.__copy_config(tmp)
					tmp._redirectlevel = self._redirectlevel + 1
					try:
						tmp.fetch()
					except Error, errmsg:
						tmp._errmsg = errmsg
					self.__dict__ = tmp.__dict__
					self.process_url()
				else:
					raise Error, "too many redirects"

		else:											### other HTTP error
			raise Error, http_status_codes.get(errcode, errmsg)


	def parse_html(self):
		''' 
		Parse HTML and populate attributes. If we want to, we can also 
		descend things like frames, images, etc here.
		'''

		parser = _MyParser(formatter.NullFormatter())
		parser.feed(self._body)
		if parser.base: self.base_url = parser.base
		self.title = parser.title
		self._links = parser.anchorlist
		self._meta = parser.meta

		if self.descend_frames > 0:
			for frame in parser.frames:
				frameref = urljoin(self.base_url, strip(frame))
				tmp = self.__class__(frameref)
				self.__copy_config(tmp)
				tmp.descend_frames = self.descend_frames - 1
				try:
					tmp.fetch()
				except Error, errmsg:
					tmp._errmsg = errmsg
				self._frames.append(tmp)

		if self.run_images:
			for image in parser.images:
				imgref = urljoin(self.base_url, strip(image))
				tmp = self.__class__(imgref)
				self.__copy_config(tmp)
				try:
					tmp.fetch()
					if tmp._hdrs.content_type[:5] != 'image': 
						raise Error, "not an image"
				except Error, errmsg:
					tmp._errmsg = errmsg
				self._images.append(tmp)

		if self.run_objects:
			for object in parser.objects:
				objref = urljoin(self.base_url, strip(object))
				tmp = self.__class__(objref)
				self.__copy_config(tmp)
				try:
					tmp.fetch()
				except Error, errmsg:
					tmp._errmsg = errmsg
				self._objects.append(tmp)


	def __repr__(self):
		if type(self._errmsg) is types.InstanceType:
			self._errmsg = self._errmsg.args[-1]

		o=[]
		p = o.append
		
		p("%s" % (self.request_url))
		if self._redirectlevel > 0:
			p("   (REDIRECTED %s times)" % (self._redirectlevel))
		if self._errmsg:
			p("   (ERROR: %s)" % (self._errmsg))
			return join(o, "\n")
		p("        Expires: %s" % (self._hdrs.get('expires', self.no_data)))
		p("  Cache-Control: %s" % (self._hdrs.get('cache-control', self.no_data)))
		p("  Last-Modified: %s" % (self._hdrs.get('last-modified', self.no_data)))
		if self._validated == 0:
			o[-1] = o[-1] + " (validation not supported)"
		elif self._validated == 1:
			o[-1] = o[-1] + " (validated)"
		p("           ETag: %s" % (self._hdrs.get('ETag', self.no_data)))
		if self._hdrs.has_key('Set-Cookie'):
			p("     Set-Cookie: %s" % (self._hdrs['Set-Cookie']))
		if self._hdrs.has_key('Set-Cookie2'):
			p("    Set-Cookie2: %s" % (self._hdrs['Set-Cookie2']))
		p(" Content-Length: %s" % (self._hdrs.get('content-length', self.no_data)))
		if self._hdrs.has_key('content-length') and \
		  self._hdrs.get('content-length', -1) != self.size:
			o[-1] = o[-1] + " (actual size: %s)" % (self.size)
		p("         Server: %s" % (self._hdrs.get('server', self.no_data)))
		p("\n")
		for image in self._images:
			p('IMAGE ' + repr(image))
		for object in self._objects:
			p('OBJECT ' + repr(object))
		for frame in self._frames:
			p('FRAME ' + repr(frame))
		return join(o, "\n")

  
class _MyParser (htmllib.HTMLParser):
	''' HTML parsing class '''

	def __init__(self, formatter, verbose=0):
		self.meta = {}
		self.frames = []
		self.alt = []
		self.images = []
		self.objects = []
		htmllib.HTMLParser.__init__(self, formatter, verbose)

	def do_meta(self, attrs):
		''' handle meta tags '''
		name, content = None, None
		for n, v in attrs:
			if n == 'name':
				name = lower(v)
			if n == 'content':
				content = cdata(v)
			if name: self.meta[name] = content

	def do_frame(self, attrs):
		''' handle frame tags '''
		src = None
		for n, v in attrs:
			if n == 'src':
				src = cdata(v)
		if src: self.frames.append(src)

	def do_iframe(self, attrs):
		''' handle iframe tags '''
		self.do_frame(attrs)

	def do_img(self, attrs):
		''' handle img tags '''
		alt, src = None, None
		for n, v in attrs:
			if n == 'alt':
				alt = cdata(v)
			if n == 'src':
				src = cdata(v)
		if alt: self.alt.append(alt)
		# very simple "cache" to avoid appending lots of the same image
		# if it's referenced from the same page.
		if src and src not in self.images: self.images.append(src)

	def start_body(self, attrs):
		''' handle body tag '''		
		bg = None
		for n, v in attrs:
			if n == 'background':
				bg = cdata(v)
		if bg: self.images.append(bg)

	def start_script(self, attrs):
		''' handle script tag '''
		script = None
		for n, v in attrs:
			if n == 'src':
				script = cdata(v)
		if script: self.objects.append(script)

	def end_script(self):
		pass

	def do_link(self, attrs):
		''' handle link tag w/ css '''
		rel, href = None, None
		for n, v in attrs:
			if n == 'rel':
				rel = cdata(v)
			if n == 'href':
				href = cdata(v)
		if rel and lower(rel) == 'stylesheet' and href: self.objects.append(href)


def handle_alarm(sig_num, stack):
	import signal
	signal.alarm(0)
	raise Error, 'timeout'
	

def cdata(raw):
	''' normalise a CDATA HTML element - this gets around a bug in htmllib. '''
	return translate(raw, cdata_tr, '\r')



